//
// chamereon USB + Logiana board , PROM Emulator Design
//
// Copyright(C) 2003 Tatusuyuki Satoh
//
// EZ-USB pin assign
// +---------+----------------+
// |PC[3..0] | RegisterOffset |
// | D[7..0] | DataBus        |
// |FWR      | WriteStrobe    |
// |FRD      | ReadStrobe     |
// +---------+----------------+
//
// write register map
// +----+---+---+---+---+---+---+---+---+----------------------------+
// |offs|b7 |b6 |b5 |b4 |b3 |b2 |b1 |b0 | Contents                   |
// +----+---+---+---+---+---+---+---+---+----------------------------+
// |+00 |WD7|WD6|WD5|WD4|WD3|WD2|WD1|WD0| Data Store                 |
// |+01 |A7 |A6 |A5 |A4 |A3 |A2 |A1 |A0 | Address L                  |
// |+02 |A15|A14|A13|A12|A11|A10|A9 |A8 | Address M                  |
// |+03 | X | X | X | X | X |A18|A17|A16| Address H & emu enable     |
// |+06 |EMU|RSE|RSL|TPM| X | X | X | X | ModeControll               |
// |+07 | X |FL |R18|R17|R16|R15|R14|R13| address enable mask        |
// +----+---+---+---+---+---+---+---+---+----------------------------+
//
// read register map
// +----+---+---+---+---+---+---+---+---+----------------------------+
// |offs|b7 |b6 |b5 |b4 |b3 |b2 |b1 |b0 |	Contents                   |
// +----+---+---+---+---+---+---+---+---+----------------------------+
// |+00 |RD7|RD6|RD5|RD4|RD3|RD2|RD1|RD0| Data ReadBack              | 
// |+04 | 0 | 0 | 0 | 0 | 0 | 0 |VCC|TRG| Sense VCC & Trigger        |
// |+0D | 0 | 1 | 0 | 1 | 0 | 0 | 1 | 0 | Fixed Code :'R'(8'h52)     |
// |+0E | 0 | 1 | 0 | 0 | 0 | 1 | 0 | 1 | Fixed Code :'E'(8'h45)     |
// |+0F |MV3|MV2|MV1|MV0|RC3|RC2|RC1|RC0| Version Code               |
// +----+---+---+---+---+---+---+---+---+----------------------------+
//
// Writre Register Spec.
// +-------+-----------------------------------------------+
// |Param. | Spec.                                         |
// +-------+-----------------------------------------------+
// |WD     | When it is written four times,                |
// |[7..0] | DWORD data is stored to emulation RAM.        |
// |       | And auto increment address register.          |
// |       | Can't access during EMU=1                     |
// +-------+-----------------------------------------------+
// |A      | RAM Address for write and read data,and Triger|
// |[18..0]| Ditect Address.                               |
// |       | A[1..0] should be set 2'b00 for write access  |
// +-------+-----------------------------------------------+
// |EMU    | 0=emulate disable, can access by host.        |
// |       | 1=emulate enable , can't access by host.      |
// +-------+-----------------------------------------------+
// |RSE    | Reset Signal                                  |
// |RSL    | RSE=0 : Reset = Hi-z                          |
// |RSL    | RSE=1 : Reset = RSL                           |
// +-------+-----------------------------------------------+
// |TPM    | Trigger Pulse Mode                            |
// |       | 0=Single Trigger(write access to clear)       |
// |       | 1=Pulse Trigger                               |
// +-------+-----------------------------------------------+
// |R13    | 1=ROM.A13 enable (128Kbit, or larger)         |
// +-------+-----------------------------------------------+
// |R14    | 1=ROM.A14 enable (256Kbit, or larger)         |
// +-------+-----------------------------------------------+
// |R15    | 1=ROM.A15 enable (512Kbit, or larger)         |
// +-------+-----------------------------------------------+
// |R16    | 1=ROM.A16 enable (1Mbit  , or larger)         |
// +-------+-----------------------------------------------+
// |R17    | 1=ROM.A17 enable (2Mbit  , or larger)         |
// +-------+-----------------------------------------------+
// |R18    | 1=ROM.A18 enable (4Mbit  , or larger)         |
// +-------+-----------------------------------------------+
// |FL     | 0=A18 location is DIP32.31 (EPROM)            |
// |       | 1=A18 location is DIP32.1  (EEPROM/Flash)     |
// +-------+-----------------------------------------------+
//
// Read Register Spec.
// +-------+-----------------------------------------------+
// |Param. | Sprc.                                         |
// +-------+-----------------------------------------------+
// |RD     | When HRD is '1',readback emulation RAM.       |
// |[7..0] | And auto increment address register.          |
// |       | Can't access during EMU=1                     |
// +-------+-----------------------------------------------+
// |VCC    | Target VCC level , 1=Avtive                   |
// +-------+-----------------------------------------------+
// |TRG    | Trigger Ditect signal                         |
// |       | TRG clear when any write access               |
// +-------+-----------------------------------------------+
// |MV,RC  | Design Version/Release code                   |
// +-------+-----------------------------------------------+
//

module top(
	osc_clk,CLK24M,
	cgate,dgate,trig,rsto,
	h_pc,h_data,h_rd,h_wr,
	rom_a,rom_d,rom_ce,rom_oe,rom_gnd,rom_vcc,rom_p1,
	ram_ce2,ram_clk,ram_gw,ram_oe,ram_adsc,ram_adv,ram_add,ram_data,
   CN2_34,CN2_35,CN2_36,CN2_37,CN2_38,CN2_39,CN2_40,CN3_29,NC1,CN3_31
);
input osc_clk , CLK24M;
output cgate,dgate,trig,rsto;
// HOST INTERFACE
input [3:0] h_pc;		// host PortC
inout [7:0] h_data;	// host data bus
input h_rd;	
input h_wr;
// TARGET PROBE
input [18:0] rom_a;
inout [7:0] rom_d;
input rom_ce,rom_oe,rom_gnd,rom_vcc,rom_p1;
// PBSRAM
output ram_ce2;
inout ram_clk;
output ram_gw;
output ram_oe;
output ram_adsc;
output ram_adv;
output [16:0] ram_add;
inout [31:0] ram_data;

//No-use
output CN2_34,CN2_35,CN2_36,CN2_37,CN2_38,CN2_39,CN2_40,NC1,CN3_29,CN3_31;

// -------------------- Internal Signal

// ROM
reg  [7:0] rom_do;		// ROM output from PBSRAM

// PBAM
reg [16:0] ram_adr_l;
reg ram_gw_l;
reg ram_oe_l;
reg ram_den;
//reg ram_adsc_l;
reg [31:0] ram_do;
reg [1:0] dsel;

// Access Trigger
reg trg_ll,trg_lh,trg_f1,trg_out;
// Reset Output
reg c4_oe,c4_out;
// Trigger
reg trg_pulse;

// -------------------- HOST INTERFACE --------------------
reg [7:0]  host_dout;
reg [18:0] host_addr;
reg f18en,r18en,r17en,r16en,r15en,r14en,r13en;
reg host_wreq; // Host WriteRequest
reg emu_run;   // Emuration run


wire h_acc = h_wr & h_rd; // Fast Write & Fast Read

// -------------------- TARGET ROM --------------------
wire [18:0] rom_addr;
wire rom_vccs , rom_oes;

// -------------------- HOST Write I/F ---------------------
reg wreq_l;

always @(negedge h_acc or posedge ram_oe_l)
begin
	if(ram_oe_l)
	begin
		// clear write request after HOST write phase 1
		host_wreq <= 0;
	end else begin
		case( {h_wr,h_pc} )
		0:begin	// WR[0] DataWrite
			case( host_addr[1:0] )
				2'b00:ram_do[ 7: 0] <= h_data; 		// increment counter
				2'b01:ram_do[15: 8] <= h_data;
				2'b10:ram_do[23:16] <= h_data;
				2'b11:begin ram_do[31:24] <= h_data; host_wreq <= 1; end
			endcase
		end
		1:begin	// WR[1] address L
			host_addr[7:0] <= h_data;
		end
		2:begin // WR[2] address M
			host_addr[15:8] <= h_data;
		end
		3:begin // WR[3] address H , select , R/W
			host_addr[18:16] <= h_data[2:0];
		end
		6:begin // WR[6] mode,reset controll
			trg_pulse        <= h_data[4]; // TriggerMode
			c4_out           <= h_data[5]; // Reset Level
			c4_oe            <= h_data[6]; // Reset Enable
			emu_run          <= h_data[7]; // -Emulation / Host read
		end
		7:begin // WR[7] rom selector
			{f18en,r18en,r17en,r16en,r15en,r14en,r13en} <= h_data[6:0];
		end
		endcase

		// increment address counter when Write & Read
		if( h_pc == 0 )
			host_addr[11:0] <= host_addr[11:0] + 1;
	end
end

// -------------------- HOST Read I/F ---------------------
assign h_data = h_rd ? 8'bzzzzzzzz : 
				h_pc==13? 8'h52 : // +0d:'R'
				h_pc==14? 8'h45 : // +0e:'E'
				h_pc==15? 8'h10 : // +0f:Version 01.00
				h_pc==4 ? {6'b000000,rom_vccs,trg_out} : // +04 trg
				rom_do;           // +00:ReadBack

// --------------------- PROM pin selector --------------------
assign rom_addr[12:0] = rom_a[12:0];
assign rom_addr[13] = r13en& rom_a[13];
assign rom_addr[14] = r14en& rom_a[14];
assign rom_addr[15] = r15en& rom_a[15];
assign rom_addr[16] = r16en& rom_a[16];
assign rom_addr[17] = r17en& rom_a[17];
assign rom_addr[18] = r18en& (f18en ? rom_p1 : rom_a[18]);

// DIP28 / DIP32 VCC select
assign rom_vccs       = r16en ? rom_vcc   : rom_a[17];

// data output enable , CS , OE , POWER
assign rom_act = ~rom_gnd && rom_vccs && emu_run;
assign rom_oes = ~rom_ce && ~rom_oe && rom_act;
assign rom_d   = rom_oes ? rom_do : 8'bzzzzzzzz;

// -------------------- CLOCK Generator --------------------
reg clk50m;
always @(posedge osc_clk) clk50m = ~clk50m;

wire sel_clk = emu_run ? clk50m : CLK24M;

// -------------------- PBSRAM interface --------------------
// access timming
always @(posedge ram_clk)
begin
  if(wreq_l)
  begin
    // HOST-WR:1
    ram_oe_l   <= 1'b1;    // negate OE 
    // HOST-WR:2
    ram_den    <= ram_oe_l;// data output enable
    // HOST-WR:3
//    ram_adsc_l <= ram_den;
    ram_gw_l   <= ~ram_den;  // assert GW (and clear host_wreq)
    //
	 wreq_l     <= ~ram_den;
  end else begin
    wreq_l <= host_wreq;
    // HOST-WR:4
//    ram_adsc_l <= 0;
    ram_den    <= 1'b0;    // data output disable
    ram_gw_l   <= 1'b1;    // negate GW & data bus floating
    // HOST-WR:5
	 ram_oe_l   <= ram_den;  // assert OE
  end

  // data latch			 
  if(~ram_oe_l)
    rom_do <= (dsel == 2'b00) ? ram_data[7:0]   :
              (dsel == 2'b01) ? ram_data[15:8]  :
              (dsel == 2'b10) ? ram_data[23:16] :
                                ram_data[31:24];

  // PBRAM Address Selector
  dsel      <= emu_run ? rom_addr[1:0]  : host_addr[1:0]; 
  ram_adr_l <= emu_run ? rom_addr[18:2] : host_addr[18:2]; 

  // Access Trigger
//  trg_lh <= host_addr[18:13]==rom_addr[18:13];
//  trg_ll <= (host_addr[12:0]==rom_addr[12:0]) & rom_oes;
  trg_lh <= host_addr[18:13]==ram_adr_l[16:11];
  trg_ll <= (host_addr[12:0]=={ram_adr_l[10:0],dsel}) & rom_oes;
  // 2clock filter
  trg_f1 <= trg_lh & trg_ll;
  // sense
  if(trg_f1 & trg_lh & trg_ll)
    trg_out <= 1;
  else if(~h_wr || trg_pulse)
    trg_out <= 0;
end

// static signals
assign ram_adsc = 1'b0;		// always enable
assign ram_adv  = 1'b0;

assign ram_ce2  = 1'b0;		// always access
assign ram_clk  = sel_clk;
assign ram_oe   = ram_oe_l;	// assert only PROM read timming
assign ram_gw   = ram_gw_l;	// assert only HOST write timming
assign ram_add  = ram_adr_l;// address bus

// PBSRAM data bus output controll
// output only GW=L cycle
//assign ram_data = (ram_oe_l & ram_den) ? ram_do : 32'bzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzz;
assign ram_data = ram_den ? ram_do : 32'bzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzz;

// Clock Output
assign cgate = ~rom_act; // Controll gate
assign dgate = ~rom_oes; // Data gate
assign trig  = trg_out;                 // AccessTrigger
assign rsto  = ~c4_oe  ? 1'bz : c4_out; // Reset Output

//No-use
assign CN2_34 = 1'b0;
assign CN2_35 = 1'b0;
assign CN2_36 = 1'b0;
assign CN2_37 = 1'b0;
assign CN2_38 = 1'b0;
assign CN2_39 = 1'b0;
assign CN2_40 = 1'b0;
//assign CN2_36 = emu_run;
//assign CN2_37 = rom_ce;
//assign CN2_38 = rom_oe;
//assign CN2_39 = rom_gnd;
//assign CN2_40 = h_acc;

assign NC1    = 1'b0;
assign CN3_29 = 1'b0;
assign CN3_31 = 1'b0;

endmodule
